// Copyright (C) 2009-2009 Martin Sandve Alnæs
// Licensed under the GNU LGPL Version 3.0.
//
// First added:  2009-01-01
// Last changed: 2009-04-01
//
// This demo program computes a projection of a function
// between two scalar finite element spaces.

#include <dolfin.h>
#include "generated_code/Projection.h"

using namespace dolfin;
using namespace Projection;

class Weight: public Function
{
public:
  Weight(const FunctionSpace & V):
    Function(V) {}

  void eval(double* values, const double* x) const
  {
    values[0] = 1.0;
  }
};

class Source: public Function
{
public:
  Source(const FunctionSpace & V):
    Function(V) {}

  void eval(double* values, const double* x) const
  {
    double dx = x[0] - 0.5;
    double dy = x[1] - 0.5;
    //values[0] = 500.0*exp(-(dx*dx + dy*dy)/0.02);
    values[0] = x[0] + x[1];
  }
};

int main()
{
  UnitSquare mesh(10, 10);

  BilinearForm::TrialSpace V(mesh);
  CoefficientSpace_c C(mesh);
  CoefficientSpace_f F(mesh);

  Weight c(C);
  Source f(F);
  Function u(V);

  BilinearForm a(V, V);
  LinearForm L(V);

  a.c = c;
  L.f = f;

  Matrix A;
  Vector b;
  Assembler::assemble(A, a);
  Assembler::assemble(b, L);

  // Display matrix and vector
  cout << endl << "A" << endl;
  A.disp();
  cout << endl << "b" << endl;
  b.disp();

  MassMatrix B(mesh);
  cout << endl << "Default MassMatrix" << endl;
  B.disp();

  cout << endl << "Solving:" << endl;
//pde.solve(u);
  LUSolver solver(symmetric);
  solver.solve(A, u.vector(), b);

  File u_file("u.pvd");
  u_file << u;
  
  // Interpolate f onto the discrete space V
  Source f2(V);
  Function fi(V);
  f2.interpolate(fi.vector(), V);
  
  // Compare projection solution with interpolation
  cout << endl << "fi" << endl;
  fi.vector().disp();
  cout << endl << "u" << endl;
  u.vector().disp();
  cout << endl;
  cout << "done" << endl;
  
  //File f_file("f.pvd");
  //f_file << fi;
  
  //plot(u);
  
  return 0;
}

